
/*
Note:
*/

#define F_CPU 16000000UL
#include "trtSettings.h"
#include "trtkernel_1284.c"
#include <stdio.h>
#include <stdlib.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <math.h>     // for sine
#include <string.h>
#include "i2cmaster.h"

#define BIT_0 0b00000001
#define BIT_1 0b00000010
#define BIT_2 0b00000100
#define BIT_3 0b00001000
#define BIT_4 0b00010000
#define BIT_5 0b00100000
#define BIT_6 0b01000000
#define BIT_7 0b10000000

#define ledOnboard BIT_2

// serial communication library
// Don't mess with the semaphores
#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE 2 // user hit <enter>
#include "trtUart.h"
#include "trtUart.c"
#include <avr/wdt.h>
// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

// input arguments to each thread
// not actually used in this example
int args[2] ;

// IMU 
#define Dev24C02  0b11010010      // device address of IMU, see datasheet
#define ACCEL_XOUT_H 0x3B
#define ACCEL_XOUT_L 0x3C
#define ACCEL_YOUT_H 0x3D
#define ACCEL_YOUT_L 0x3E
#define ACCEL_ZOUT_H 0x3F
#define ACCEL_ZOUT_L 0x40
#define TEMP_OUT_H R 0x41
#define TEMP_OUT_L R 0x42
#define GYRO_XOUT_H  0x43
#define GYRO_XOUT_L  0x44
#define GYRO_YOUT_H  0x45
#define GYRO_YOUT_L  0x46
#define GYRO_ZOUT_H  0x47
#define GYRO_ZOUT_L  0x48
#define PWR_MGMT_1   0x6B
#define WHO_AM_I	 0x75

#define RA_SMPLRT_DIV		0x19
#define RA_CONFIG			0x1A
#define RA_GYRO_CONFIG		0x1B
#define RA_ACCEL_CONFIG		0x1C
#define RA_FF_THR			0x1D
#define RA_FF_DUR			0x1E
#define RA_MOT_THR			0x1F
#define RA_MOT_DUR			0x20
#define RA_ZRMOT_THR		0x21
#define RA_ZRMOT_DUR		0x22
#define RA_FIFO_EN			0x23

#define readSensor(regH) ((imuRead(regH) << 8) | imuRead(regH+1))

// --- TEST VARIABLES	  ---------------------------------

int xAccel = 0;

// --- Init Functions -------------------------
void imuInit(){
	imuWrite(PWR_MGMT_1, 0x00);		// Wake up imu from sleep mode
}

void initADC(void) {
	// init ADC and start first conversion
	DDRA = 0;
	ADMUX = (1 << ADLAR) | (1 << REFS0);
	ADCSRA = ((1 << ADEN) | (1 << ADSC)) + 7;
}

// --- IMU Read/Write Methods  -----------------------------
unsigned char imuRead(unsigned char reg){
	unsigned char ret;
	i2c_start_wait(Dev24C02+I2C_WRITE);
	i2c_write(reg);
	i2c_rep_start(Dev24C02+I2C_READ);
	ret = i2c_readNak();
	i2c_stop();
	
	return ret;
}

void imuWrite(unsigned char reg, unsigned char value){
	i2c_start_wait(Dev24C02+I2C_WRITE);			// set device address and write mode
	i2c_write(reg);
	i2c_write(value);
	i2c_stop();
}

// --- IMU Functions  ----------------------------------------
void imuReadAccelGyro(int* sensorArr) {	
	sensorArr[0] = readSensor(ACCEL_XOUT_H);
	sensorArr[1] = readSensor(ACCEL_YOUT_H);
	sensorArr[2] = readSensor(ACCEL_ZOUT_H);
	sensorArr[3] = readSensor(GYRO_XOUT_H);
	sensorArr[4] = readSensor(GYRO_YOUT_H);
	sensorArr[5] = readSensor(GYRO_ZOUT_H);
}

// --- ADC Functions -----------------------------
void readADC(int* array) {
	int i;
	for (i = 0; i < 5; i++) {
		// Read sample
		while (ADCSRA & (1 << ADSC));
		array[i] = ADCH;
		// Move to next port and start next conversion
		if (i == 4) ADMUX = (1 << ADLAR) | (1 << REFS0);
		else 		ADMUX = ((1 << ADLAR) | (1 << REFS0)) + i + 1;
		ADCSRA |= (1<<ADSC);
	}
}

// --- Touch sensors --------------------------------------
void readTouch(int* array) {
	array[0] = PINB & BIT_0;
	array[1] = PINB & BIT_1;
}

// --- Read and Send Sensors ----------------------------------------
void prepareMsg(char *msg, unsigned int *sensors) {
	msg[0] = 0xA1;
	msg[1] = 0xB2;
	msg[2] = 0xC3;
	msg[3] = 0xD4;
	
	for (int i = 0; i < 11; i++) {
		msg[2*i + 4] = sensors[i];
		msg[2*i + 5] = sensors[i] >> 8;
	}
	
	msg[4 + 22] = sensors[11];
	msg[4 + 22 + 1] = sensors[12];
	
	// Checksum
	unsigned int sum = 0;
	for (int i = 0; i < 24; i++) {
		sum += msg[4 + i];
	}
	msg[4 + 24] = sum;
	msg[4 + 24 + 1] = sum >> 8;
}

void readAndSendSensors(void* args) {
	uint32_t interval, nextSend;
	char msg[4 + 22 + 2 + 2];	// Header, data, checksum
	int accelGyroSensors[16][6];
	unsigned int flexSensors[16][5];
	unsigned int touchSensors[16][2];
	unsigned int sensorsFiltered[13];
	
	interval = SECONDS2TICKS(.03);
	
	i2c_init();
	imuInit();
	initADC();
	
	DDRB = 0;
	PORTB = 3;
	
	while (1) {
		wdt_reset();
		
		// Get 16 readings (still under the time contraint for 33 Hz)
		for (int i = 0; i < 16; i++) {
			// Read IMU data 
			imuReadAccelGyro(&accelGyroSensors[i]);

			// Read flex sensors
			readADC(&flexSensors[i]);
			
			// Read touch sensors
			readTouch(&touchSensors[i]);
		}

		// While not time to send
		while (trtCurrentTime() < nextSend) ;
		
		nextSend = trtCurrentTime() + interval;
		
		// Concatenate data
		for (int i = 0; i < 6; i++) {
			sensorsFiltered[i] = 0;
			for (int j = 0; j < 16; j++) {
				sensorsFiltered[i] += accelGyroSensors[j][i] >> 8;
			}
		}
		for (int i = 0; i < 5; i++) {
			sensorsFiltered[6+i] = 0;
			for (int j = 0; j < 16; j++) {
				sensorsFiltered[6+i] += flexSensors[j][i];
			}
		}
		for (int i = 0; i < 2; i++) {
			sensorsFiltered[11+i] = 0;
			for (int j = 0; j < 16; j++) {
				sensorsFiltered[11+i] += touchSensors[j][i];
			}
		}
			
			
		// Send data encoded
		prepareMsg(msg, sensorsFiltered);
		fwrite(msg, 4 + 22 + 2 + 2, 1, stdout);
		
		// Print info
		// fprintf(stdout, "Num readings: %d\n", sIndex);
		// fprintf(stdout, "%d %d %d\n", sensorsFiltered[0], sensorsFiltered[1], sensorsFiltered[2]);
		// fprintf(stdout, "%d %d %d\n", sensorsFiltered[3], sensorsFiltered[4], sensorsFiltered[5]);
		// fprintf(stdout, "%u %u %u %u %u\n", sensorsFiltered[6], sensorsFiltered[7], sensorsFiltered[8], sensorsFiltered[9], sensorsFiltered[10]);		
		
		
		
		
		
		
		
		
		
		// Read IMU data 
		// imuReadAccelGyro(&sensors[0]);

		// Read flex sensors
		// readADC(&sensors[6]);
		// readADC(flex);
				
		// Send data encoded
		// prepareMsg(msg, sensors, 11);
		// fwrite(msg, 4 + 22, 1, stdout);
		
		// Sleep 
		// t = tStart + interval;
		// trtSleepUntil(t, t+interval);
	}
}

// --- Test Program ----------------------------------------
// void testSendSensors(void* args) {
	// uint32_t t, interval;
	// char msg[26];
	// unsigned int sensors[] = {-1, 0, 1, -30000, 0, 30000, 1, 2, 3, 4, 5};
	
	// interval = SECONDS2TICKS(.03);
	
	// while (1) {
		// prepareMsg(msg, sensors);
		// fprintf(stdout, msg);
		// fwrite(msg, 4 + 22, 1, stdout);
		
		// Sleep 
		// t = trtCurrentTime() + interval;
		// trtSleepUntil(t, t+interval);
	// }
// }

void testSerialOut(void* args) {
  uint32_t t, interval;
  
  interval = SECONDS2TICKS(.5);
  
  while (1) {
	// PORTD ^= BIT_2;
	
    // comand: "<parameter> <value>"
    fprintf(stdout, "Testing Serial\n");
	
	// Sleep
    t = trtCurrentTime() + interval;
    trtSleepUntil(t, t + interval);
  }
}

void testSerialIn(void* args) {
  char param[4];
  double value;
    
  while (1) {	
	PORTD ^= BIT_2;
	
    // comand: "<parameter> <value>"
	fprintf(stdout, ">");
    fscanf(stdin, "%s%f", param, &value);
	fprintf(stdout, "Got: %s %f\n", param, value);
	

    // TODO: NEEDED?
    //trtWait(SEM_STRING_DONE);
  }
}

void testLED(void) {
	for (int i = 0; i < 10; i++) {
		PORTD ^= BIT_2;
		_delay_ms(100);
	}
}

void blinkLED(void) {
	PORTD ^= BIT_2;
	_delay_ms(100);
	PORTD ^= BIT_2;
}

// --- Main Program ----------------------------------------
int main(void) {

	// Watchdog Timer
	cli();
	wdt_reset(); //__watchdog_reset();
	WDTCSR |= (1 << WDCE) | (1 << WDE);
	WDTCSR = (1 << WDE) | (1 << WDP2) | (1 << WDP0);
	sei();

	// LED
	DDRD |= BIT_2;
	PORTD |= BIT_2;
	blinkLED();
	
	//init the UART -- trt_uart_init() is in trtUart.c
	trt_uart_init();
	stdout = stdin = stderr = &uart_str;
	fprintf(stdout,"\n\r TRT Starting\n\r\n\r");
	
	// start TRT
	trtInitKernel(80); // 80 bytes for the idle task stack
		
	// --- create semaphores ----------
	// You must creat the first two semaphores if you use the uart
	trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0) ; // uart receive ISR semaphore
	trtCreateSemaphore(SEM_STRING_DONE, 0) ;  // user typed <enter>
	
	// Signals and variable protection
	
	// --- test tasks  ----------------
	// trtCreateTask(testSerialOut, 100, SECONDS2TICKS(0.0), SECONDS2TICKS(0.0), &(args[0]));
	// trtCreateTask(testSerialIn, 100, SECONDS2TICKS(0.0), SECONDS2TICKS(0.0), &(args[0]));
	// trtCreateTask(runMotorPrintSpeed, 100, SECONDS2TICKS(0.0), SECONDS2TICKS(0.0), &(args[0]));

	
	// --- create tasks  ----------------
	// trtCreateTask(imuPoll, 1000, SECONDS2TICKS(0.0), SECONDS2TICKS(0.0), &(args[0]));
	// trtCreateTask(testSendSensors, 1000, SECONDS2TICKS(0.0), SECONDS2TICKS(0.0), &(args[0]));
	trtCreateTask(readAndSendSensors, 2000, SECONDS2TICKS(0.0), SECONDS2TICKS(0.0), &(args[0]));
		
	// --- Idle task --------------------------------------
	// just sleeps the cpu to save power 
	// every time it executes
	set_sleep_mode(SLEEP_MODE_IDLE);
	sleep_enable();
	while (1) {
		sleep_cpu();
	}
}
